@@ -1,6 +1,4 @@
 """Wiring module."""
-
-import asyncio
 import functools
 import inspect
 import importlib
@@ -58,7 +56,6 @@ except ImportError:
 
 from . import providers
 
-
 if sys.version_info[:2] == (3, 5):
     warnings.warn(
         "Dependency Injector will drop support of Python 3.5 after Jan 1st of 2022. "
@@ -321,7 +318,7 @@ class InspectFilter:
     def _is_starlette_request_cls(self, instance: object) -> bool:
         return starlette \
                and isinstance(instance, type) \
-               and issubclass(instance, starlette.requests.Request)
+               and _safe_is_subclass(instance, starlette.requests.Request)
 
     def _is_builtin(self, instance: object) -> bool:
         return inspect.isbuiltin(instance)
@@ -600,71 +597,6 @@ def _get_patched(fn, reference_injections, reference_closing):
     return patched
 
 
-def _get_sync_patched(fn):
-    @functools.wraps(fn)
-    def _patched(*args, **kwargs):
-        to_inject = kwargs.copy()
-        for injection, provider in _patched.__injections__.items():
-            if injection not in kwargs \
-                    or _is_fastapi_default_arg_injection(injection, kwargs):
-                to_inject[injection] = provider()
-
-        result = fn(*args, **to_inject)
-
-        for injection, provider in _patched.__closing__.items():
-            if injection in kwargs \
-                    and not _is_fastapi_default_arg_injection(injection, kwargs):
-                continue
-            if not isinstance(provider, providers.Resource):
-                continue
-            provider.shutdown()
-
-        return result
-    return _patched
-
-
-def _get_async_patched(fn):
-    @functools.wraps(fn)
-    async def _patched(*args, **kwargs):
-        to_inject = kwargs.copy()
-        to_inject_await = []
-        to_close_await = []
-        for injection, provider in _patched.__injections__.items():
-            if injection not in kwargs \
-                    or _is_fastapi_default_arg_injection(injection, kwargs):
-                provide = provider()
-                if inspect.isawaitable(provide):
-                    to_inject_await.append((injection, provide))
-                else:
-                    to_inject[injection] = provide
-
-        async_to_inject = await asyncio.gather(*[provide for _, provide in to_inject_await])
-        for provide, (injection, _) in zip(async_to_inject, to_inject_await):
-            to_inject[injection] = provide
-
-        result = await fn(*args, **to_inject)
-
-        for injection, provider in _patched.__closing__.items():
-            if injection in kwargs \
-                    and not _is_fastapi_default_arg_injection(injection, kwargs):
-                continue
-            if not isinstance(provider, providers.Resource):
-                continue
-            shutdown = provider.shutdown()
-            if inspect.isawaitable(shutdown):
-                to_close_await.append(shutdown)
-
-        await asyncio.gather(*to_close_await)
-
-        return result
-    return _patched
-
-
-def _is_fastapi_default_arg_injection(injection, kwargs):
-    """Check if injection is FastAPI injection of the default argument."""
-    return injection in kwargs and isinstance(kwargs[injection], _Marker)
-
-
 def _is_fastapi_depends(param: Any) -> bool:
     return fastapi and isinstance(param, fastapi.params.Depends)
 
@@ -679,6 +611,13 @@ def _is_declarative_container(instance: Any) -> bool:
             and getattr(instance, "declarative_parent", None) is None)
 
 
+def _safe_is_subclass(instance: Any, cls: Type) -> bool:
+    try:
+        return issubclass(instance, cls)
+    except TypeError:
+        return False
+
+
 class Modifier:
 
     def modify(
@@ -821,6 +760,8 @@ class ClassGetItemMeta(GenericMeta):
 
 class _Marker(Generic[T], metaclass=ClassGetItemMeta):
 
+    __IS_MARKER__ = True
+
     def __init__(
             self,
             provider: Union[providers.Provider, Container, str],
@@ -951,3 +892,22 @@ def is_loader_installed() -> bool:
 _patched_registry = PatchedRegistry()
 _inspect_filter = InspectFilter()
 _loader = AutoLoader()
+
+# Optimizations
+from ._cwiring import _get_sync_patched  # noqa
+from ._cwiring import _async_inject  # noqa
+
+
+# Wiring uses the following Python wrapper because there is
+# no possibility to compile a first-type citizen coroutine in Cython.
+def _get_async_patched(fn):
+    @functools.wraps(fn)
+    async def _patched(*args, **kwargs):
+        return await _async_inject(
+            fn,
+            args,
+            kwargs,
+            _patched.__injections__,
+            _patched.__closing__,
+        )
+    return _patched
